#!/usr/bin/env python

"""
Translate the AST of a Python program into a more interpretable representation.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import micropython.inspect
from micropython.rsvp import *
from micropython.common import *
import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class TranslateError(ProcessingError): pass

class Label:

    "A reference to a location."

    def __init__(self, number, location=None):
        self.number = number
        self.location = location

    def __repr__(self):
        return "Label(%r, location=%r)" % (self.number, self.location)

# Program visitors.

class Translation(ASTVisitor):

    "A translated module."

    def __init__(self, module, objtable, paramtable):

        """
        Initialise the translation with an inspected 'module' and an attribute
        table 'objtable' and parameter table 'paramtable'.
        """

        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module
        self.objtable = objtable
        self.paramtable = paramtable
        self.unit = None

        # Wiring within the code.

        self.labels = {}
        self.label_number = 0
        self.loop_labels = []

        # The code itself.

        self.code = None

    def get_module_code(self):

        "Return the top-level module code."

        self.unit = self.module
        self.code = []
        self.dispatch(self.module.module)
        return self.code

    def get_code(self, unit):

        "Return the code for the given 'unit'."

        self.unit = unit
        self.code = []
        if unit.node is not None:
            self.dispatch(unit.node)
        return self.code

    def __repr__(self):
        return "Translation(%r)" % self.module

    def get_scope(self, name):
        if self.unit.has_key(name):
            return "local"
        elif self.module.has_key(name):
            return "global"
        else:
            return "builtins"

    # Code writing methods.

    def new_label(self):
        number = self.label_number
        label = Label(number)
        self.labels[label] = label
        self.label_number += 1
        return label

    def set_label(self, label):

        """
        Set the location of 'label' to that within the entire image: the
        location within the code combined with location of the code unit.
        """

        label.location = len(self.code) + self.unit.code_location

    def new_op(self, op):
        self.code.append(op)

    def replace_op(self, op):
        self.code[-1] = op

    def last_op(self):
        return self.code[-1]

    # Visitor methods.

    def default(self, node, *args):
        raise TranslateError(self.module.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def _visitAttr(self, node, classes):
        AttrInstruction, AttrIndexInstruction = classes
        self.dispatch(node.expr)

        # NOTE: Only simple cases are used for optimisations.

        last = self.last_op()
        if isinstance(last, (LoadName, LoadAttr)) and last.attr.assignments == 1:

            # Get the details of the access.

            target = last.attr.value
            target_name = target.full_name()
            table_entry = self.objtable.table[target_name]
            pos = table_entry[node.attrname]
            self.replace_op(AttrInstruction(pos))

        else:
            index = self.objtable.get_index(node.attrname)
            self.new_op(AttrIndexInstruction(index))

    def _visitName(self, node, classes):
        name = node.name
        scope = self.get_scope(name)
        #print self.module.name, node.lineno, name, scope
        NameInstruction, AttrInstruction = classes

        if scope == "local":
            unit = self.unit
            if isinstance(unit, micropython.inspect.Function):
                self.new_op(NameInstruction(unit.all_locals()[name]))
            elif isinstance(unit, micropython.inspect.Class):
                self.new_op(AttrInstruction(unit.all_class_attributes()[name]))
            elif isinstance(unit, micropython.inspect.Module):
                self.new_op(AttrInstruction(unit.module_attributes()[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Program unit %r has no local %r" % (unit, name))

        elif scope == "global":
            globals = self.module.module_attributes()
            if globals.has_key(name):
                self.new_op(AttrInstruction(globals[name]))
            else:
                raise TranslateError(self.module.full_name(), node, "Module %r has no attribute %r" % (self.module, name))

        else:
            builtins = micropython.inspect.builtins.module_attributes()
            self.new_op(AttrInstruction(builtins[name]))

    def visitAdd(self, node):

        """
        _t1 = node.left
        _t2 = node.right
        try:
            _t1.__add__(_t2)
        except AttributeError:
            _t2.__radd__(_t1)
        """

    def visitAnd(self, node): pass

    def visitAssert(self, node): pass

    def visitAssign(self, node):
        self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)

    def visitAssAttr(self, node):
        self._visitAttr(node, (StoreAttr, StoreAttrIndex))

    def visitAssList(self, node): pass

    def visitAssName(self, node):
        self._visitName(node, (StoreName, StoreAttr))

    visitAssTuple = visitAssList

    def visitAugAssign(self, node): pass

    def visitBackquote(self, node): pass

    def visitBitand(self, node): pass

    def visitBitor(self, node): pass

    def visitBitxor(self, node): pass

    def visitBreak(self, node):
        next_label, exit_label = self.loop_labels[-1]
        self.new_op(Jump(exit_label))

    def visitCallFunc(self, node):

        """
        Evaluate positional arguments, evaluate and store keyword arguments in
        the correct location, then invoke the function.
        """

        # Record the location of the invocation.

        self.new_op(MakeFrame()) # records the start of the frame

        # Evaluate the target.

        self.dispatch(node.node)

        # NOTE: Only simple cases are used for optimisations.

        last = self.last_op()
        if isinstance(last, (LoadName, LoadAttr)) and last.attr.assignments == 1:
            target = last.attr.value
        else:
            target = None

        # Evaluate the arguments.

        positional = 1

        for i, arg in enumerate(node.args):
            if isinstance(arg, compiler.ast.Keyword):
                if positional:
                    self.new_op(ReserveFrame(len(node.args) - i))
                    positional = 0

                self.dispatch(arg.expr)

                # Optimise where the target is known now.

                if target is not None:

                    # Find the parameter table entry for the target.

                    target_name = target.full_name()

                    # Look for a callable with the precise target name.

                    try:
                        table_entry = self.paramtable.table[target_name]

                    # Where no callable is present, check to see if it is a
                    # class name and find the initialiser instead.

                    except KeyError:
                        if self.objtable.table.has_key(target_name):
                            table_entry = self.paramtable.table[target_name + ".__init__"]
                        else:
                            raise

                    pos = table_entry[arg.name]
                    self.new_op(StoreFrame(pos))

                # Otherwise, generate the code needed to obtain the details of
                # the parameter location.

                else:

                    # Combine the target details with the name to get the location.
                    # See the access method on the List class.

                    try:
                        paramindex = self.paramtable.get_index(arg.name)
                    except ValueError:
                        raise TranslateError(self.module.full_name(), node, "No parameter definition exists for %r." % arg.name)

                    self.new_op(StoreFrameIndex(paramindex))

                    # use (callable+0)+paramindex+table
                    # checks embedded offset against (callable+0)
                    # moves the top of stack to frame+position

            else:
                self.dispatch(arg)

        self.new_op(LoadCallable()) # uses the start of the frame to get the callable
        self.new_op(Jump())

        # NOTE: Exception handling required.

        self.new_op(DropFrame())

    def visitClass(self, node):
        unit = self.unit
        self.unit = node.unit
        self.unit.code_location = self.module.code_location + len(self.code)
        self.dispatch(node.code)
        self.unit = unit

    def visitCompare(self, node):

        """
        self.dispatch(node.expr)
        for op_name, next_node in compare.ops:
            methods = self.comparison_methods[op_name]
            if methods is not None:
                # Generate method call using evaluated argument and next node.
            else:
                # Deal with the special operators.
            # Provide short-circuiting.
        """

    def visitConst(self, node):
        const = self.module.constant_values[node.value]
        self.new_op(LoadConst(const))

    def visitContinue(self, node):
        next_label, exit_label = self.loop_labels[-1]
        self.new_op(Jump(next_label))

    def visitDecorators(self, node): pass

    def visitDict(self, node): pass

    def visitDiscard(self, node):
        self.dispatch(node.expr)

    def visitDiv(self, node): pass

    def visitEllipsis(self, node): pass

    def visitExec(self, node): pass

    def visitExpression(self, node): pass

    def visitFloorDiv(self, node): pass

    def visitFor(self, node): pass

    def visitFrom(self, node): pass

    def visitFunction(self, node):

        # Only store the name when visiting this node from outside.

        if self.unit is not node.unit:
            self.new_op(LoadConst(node.unit))
            self._visitName(node, (StoreName, StoreAttr))

        # Visiting of the code occurs when get_code is invoked on this node.

        else:
            self.dispatch(node.code)
            self.new_op(Return())

    def visitGenExpr(self, node): pass

    def visitGenExprFor(self, node): pass

    def visitGenExprIf(self, node): pass

    def visitGenExprInner(self, node): pass

    def visitGetattr(self, node):
        self._visitAttr(node, (LoadAttr, LoadAttrIndex))

    def visitGlobal(self, node): pass

    def visitIf(self, node):
        first = 1
        exit_label = self.new_label()

        for test, body in node.tests + [(None, node.else_)]:
            if body is None:
                break
            if not first:
                self.set_label(next_label)
            if test is not None:
                self.dispatch(test)
                next_label = self.new_label()
                self.new_op(JumpIfFalse(next_label))
            self.dispatch(body)
            self.new_op(Jump(exit_label))
            first = 0

        self.set_label(exit_label)

    def visitImport(self, node): pass

    def visitInvert(self, node): pass

    def visitKeyword(self, node): pass

    def visitLambda(self, node): pass

    def visitLeftShift(self, node): pass

    def visitList(self, node): pass

    def visitListComp(self, node): pass

    def visitListCompFor(self, node): pass

    def visitListCompIf(self, node): pass

    def visitMod(self, node): pass

    def visitModule(self, node):
        self.dispatch(node.node)

    def visitMul(self, node): pass

    def visitName(self, node):
        self._visitName(node, (LoadName, LoadAttr))

    def visitNot(self, node): pass

    def visitOr(self, node): pass

    def visitPass(self, node): pass

    def visitPower(self, node): pass

    def visitPrint(self, node): pass

    def visitPrintnl(self, node): pass

    def visitRaise(self, node): pass

    def visitReturn(self, node):
        if node.value is not None:
            self.dispatch(node.value)
        self.new_op(Return())

    def visitRightShift(self, node): pass

    def visitSlice(self, node): pass

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)

    def visitSub(self, node): pass

    def visitSubscript(self, node): pass

    def visitTryExcept(self, node): pass

    def visitTryFinally(self, node): pass

    def visitTuple(self, node): pass

    def visitUnaryAdd(self, node): pass

    def visitUnarySub(self, node): pass

    def visitWhile(self, node):
        exit_label = self.new_label()
        self.dispatch(node.test)
        self.new_op(JumpIfFalse(exit_label))

        next_label = self.new_label()
        self.set_label(next_label)
        self.loop_labels.append((next_label, exit_label))

        self.dispatch(node.body)
        self.new_op(Jump(next_label))

        if node.else_ is not None:
            self.dispatch(node.else_)

        self.set_label(exit_label)
        self.loop_labels.pop()

    def visitWith(self, node): pass

    def visitYield(self, node): pass

    # Useful data.

    comparison_methods = {
        "==" : ("__eq__", "__ne__"),
        "!=" : ("__ne__", "__eq__"),
        "<" : ("__lt__", "__gt__"),
        "<=" : ("__le__", "__ge__"),
        ">=" : ("__ge__", "__le__"),
        ">" : ("__gt__", "__lt__"),
        "is" : None,
        "is not" : None,
        "in" : None,
        "not in" : None
        }

# vim: tabstop=4 expandtab shiftwidth=4
